"""Package for fitting regexes to strings."""

from metasynth.distribution.legacy.regex.base import (RegexDistribution,
                                                      UniqueRegexDistribution)

__all__ = ["RegexDistribution", "UniqueRegexDistribution"]
